import 'dart:async';
import 'dart:io';
import 'dart:typed_data';
import 'package:flutter/rendering.dart';
import 'dart:ui' as ui;
import 'package:package_info/package_info.dart';
import 'package:flutter_email_sender/flutter_email_sender.dart';
import 'package:flutter/material.dart';


/// Log to LogView
void LVFATAL(Object? m)  => _LogHistory.instance.fatal(m.toString());
/// Log to LogView
void LVERROR(Object? m)  => _LogHistory.instance.error(m.toString());
/// Log to LogView
void LVWARN(Object? m)   => _LogHistory.instance.warn(m.toString());
/// Log to LogView
void LVINFO(Object? m)   => _LogHistory.instance.info(m.toString());
/// Log to LogView
void LVLOG(Object? m)    => _LogHistory.instance.info(m.toString());
/// Log to LogView
void LVDEBUG(Object? m)  => _LogHistory.instance.debug(m.toString());
/// Log to LogView
void LVTRACE(Object? m)  => _LogHistory.instance.trace(m.toString());

/// Widget to show app log in a view and send by email to analize
class LogView {
  static LogView? instance;
  LogView(Widget child, {String? logEmail,bool showLogButton=true, void Function(FlutterErrorDetails error)? onFlutterError, void Function(dynamic error, StackTrace stackTrace)? onZoneError}) {
    instance = this;
    /*ErrorWidget.builder = (FlutterErrorDetails details) => _LogView(
      Container(
        color: Colors.red[400],
        child: ListView(children: <Widget>[Text(details.exceptionAsString()+'\n\n'+details.stack.toString(), style: TextStyle(color:Colors.white))])
      ),
      logEmail
    );*/
    
    FlutterError.onError = (FlutterErrorDetails details) async {
      if(onFlutterError !=null)
        try {
          onFlutterError(details);
        } catch(e) { print(e); }
      else {
        print(details.exceptionAsString());
        LVERROR(details.exceptionAsString());
        print(details.stack.toString());
        LVERROR(details.stack.toString());
      }
    };

    runZonedGuarded<Future<void>>(() async {          
      try {
        runApp(_LogView(child,logEmail,showLogButton));
      } catch(e) { print(e); }
    }, (dynamic error, StackTrace stackTrace) {
      if(onZoneError != null)
        try {
          onZoneError(error,stackTrace);
        } catch(e) { print(e); }
      else {
        print(stackTrace.toString());
        LVLOG(stackTrace.toString());
      }
    });
  }

  List<LogTrace> logHistory() {
    return _LogHistory.instance.messages;
  }

  static Future<String> getPackageInfo() async {
    PackageInfo packageInfo = await PackageInfo.fromPlatform();

    return '${packageInfo.appName}(${packageInfo.packageName}) ${Platform.isAndroid ? "Android" : "iOS"} ${packageInfo.version}+${packageInfo.buildNumber}';
  }

  Future<String> textReport() async {    
    String res = 'App report ${await getPackageInfo()}\n\n';

    _LogHistory.instance.messages.forEach((entry) {
      res += '${entry.message}\n';
    });

    return res;
  }

  Future<String> htmlReport() async {
    String res = '''
      <html>
        <style>
          * { font-family: "Courier New", Courier, monospace"; }
          .f { background-color: #c70000; color: white}
          .e { background-color: #7d0a0a; color: white}
          .w { background-color: #7d3f0a; color: white}
          .i { background-color: #353535; color: white}
          .d { background-color: #353535; color: #c5c5c5}
          .t { background-color: #353535; color: #848484}
        </style>
      <body>''';

    res +='<h1>App report ${await getPackageInfo()}</h1>';

    res += '<table>';
    _LogHistory.instance.messages.forEach((entry) {
      var levelClass = 't';
      switch(entry.level) {
        case _LEVELS.FATAL: levelClass = 'f'; break; 
        case _LEVELS.ERROR: levelClass = 'e'; break; 
        case _LEVELS.WARN:  levelClass = 'w'; break; 
        case _LEVELS.INFO:  levelClass = 'i'; break; 
        case _LEVELS.DEBUG: levelClass = 'd'; break; 
        case _LEVELS.TRACE: levelClass = 't'; break; 
      }

      res += '<tr><td class="$levelClass">';
      res += "${entry.message.replaceAll('<', '&lt;').replaceAll('>', '&gt;')}";
      res += '</td></tr>';
    });

    res += '</table></body></html>';

    return res;
    
  }

}

class _LogView extends StatefulWidget {
  final Widget child;
  final String? logEmail;
  final bool showLogButton;

  _LogView(this.child,this.logEmail,this.showLogButton);

  @override
  _LogViewState createState() => new _LogViewState();
}
class _LogViewState extends State<_LogView> {
  static const traceStyle = TextStyle(color: Colors.grey, fontFamily: 'Courier', fontSize: 12);
  static const debugStyle = TextStyle(color: Color(0xFFC9C9C9), fontFamily: 'Courier', fontSize: 12);
  static const normalStyle = TextStyle(color: Colors.white, fontFamily: 'Courier', fontSize: 12);
  static const warnStyle = TextStyle(color: Colors.yellow, fontFamily: 'Courier', fontSize: 12);
  static const errorStyle = TextStyle(color: Colors.red, fontFamily: 'Courier', fontSize: 12);
  static const fatalStyle = TextStyle(color: Color(0xFFB71C1C), fontFamily: 'Courier', fontSize: 12);

  final GlobalKey repaintKey = GlobalKey();

  bool expanded = false;
  _LEVELS? level = _LEVELS.TRACE;
  String? filter;

  @override
  Widget build(BuildContext context) {
    return Stack(textDirection: TextDirection.ltr, children: [
      if(!expanded)
        RepaintBoundary(child:widget.child),
      if(widget.showLogButton)
        Directionality(textDirection: TextDirection.ltr,child:
          (expanded) ?
            getLog()
          :
            Align(alignment: Alignment.centerRight,child: 
              Padding(padding:EdgeInsets.only(bottom:40,right:20), child: GestureDetector(
                child: Icon(Icons.message,color: Colors.black.withOpacity(0.04),),
                onLongPress: () => setState(() => expanded = true),
              )
            )          
        ))
    ]);
    
  }

  Widget getLog() {
    return MaterialApp(
      debugShowCheckedModeBanner: false,
      home: Scaffold(body:Container(
        color: Colors.black87,
        width: double.infinity,
        height: double.infinity,
        child: SafeArea(bottom:true,child:Column(children: [
          Expanded(child:          
            Stack(children: [
              ListView(children: <Widget>[
                for(var message in _LogHistory.instance.messages)
                  if(message.level <= level && (filter == null || filter!.isEmpty || message.message.toLowerCase().contains(filter!.toLowerCase())))                  
                    if(message.level == _LEVELS.FATAL)
                      SelectableText(message.message, style: fatalStyle)
                    else if(message.level == _LEVELS.ERROR)
                      SelectableText(message.message, style: errorStyle)
                    else if(message.level == _LEVELS.WARN)
                      SelectableText(message.message, style: warnStyle,)
                    else if(message.level == _LEVELS.TRACE)
                      SelectableText(message.message, style: traceStyle,)
                    else if(message.level == _LEVELS.DEBUG)
                      SelectableText(message.message, style: debugStyle,)
                    else
                      SelectableText(message.message, style: normalStyle,),                   
              ]),
              Align(alignment: Alignment.topRight,child: IconButton(icon: Icon(Icons.close, color: Colors.white), onPressed: onClose,),)
            ])
          ),
          Container(
            padding: EdgeInsets.symmetric(horizontal:5),
            color: Colors.black54,
            child: Row(mainAxisAlignment: MainAxisAlignment.spaceAround, children: [
              Expanded(child:TextField(
                cursorColor: Colors.white,                            
                decoration: InputDecoration(                
                  border: InputBorder.none,
                  focusedBorder: InputBorder.none,
                  enabledBorder: InputBorder.none,
                  errorBorder: InputBorder.none,
                  disabledBorder: InputBorder.none,     
                  hintText: 'filter',
                  hintStyle: TextStyle(color:Colors.grey)           
                ),            
                style: TextStyle(color:Colors.white),
                onChanged: onFilter,
              )),
              Padding(
                padding: EdgeInsets.symmetric(horizontal: 10),
                child:DropdownButton<_LEVELS>(
                  dropdownColor: Colors.black87,
                  value: level,
                  iconSize: 24,
                  elevation: 16,
                  style: TextStyle(color: Colors.white),
                  
                  onChanged: (_LEVELS? newValue) {
                    setState(() {
                      level = newValue;
                    });
                  },
                  items: <_LEVELS>[_LEVELS.FATAL,_LEVELS.ERROR,_LEVELS.WARN,_LEVELS.INFO,_LEVELS.DEBUG,_LEVELS.TRACE]
                      .map<DropdownMenuItem<_LEVELS>>((_LEVELS value) {
                    return DropdownMenuItem<_LEVELS>(
                      value: value,
                      child: Text(value.ToString()),
                    );
                  }).toList(),
                ),
              ),
              ElevatedButton(
                style: ButtonStyle(
                  backgroundColor: MaterialStateProperty.all(Colors.black87)
                ),
                onPressed: onSend,
                child: Text('Send', style: TextStyle(color: Colors.white),),
              ),            
            ])
          )
        ])
      )))
    );
  }

  void onClose() {
    setState(() => expanded = false);
  }


  void onFilter(String text) {
    setState(() => filter = text);
  }


  void onSend() {
    
    try {
      onClose();

      (() async {
        PackageInfo packageInfo = await PackageInfo.fromPlatform();
        String body = '${await LogView.getPackageInfo()}\n\n';
        
        for(var message in _LogHistory.instance.messages)
          body += '\n${message.message}';
                  
        final Email email = Email(
          body: body,
          subject: '${packageInfo.appName}(${packageInfo.packageName}) error log',
          recipients: widget.logEmail != null? [widget.logEmail!] : [],      
          isHTML: false,
        );

        await FlutterEmailSender.send(email);
      })();
    }catch (e) {
      LVERROR(e);
    }

  }

  // return a png of screen image
  Future<ByteData?> getScreenShot() async {
    RenderRepaintBoundary boundary = repaintKey.currentContext!.findRenderObject() as RenderRepaintBoundary;
    ui.Image image = await boundary.toImage();
    return await image.toByteData(format: ui.ImageByteFormat.png);    
  }
}


enum _LEVELS { FATAL,ERROR,WARN,INFO,DEBUG,TRACE }

extension _LevelsExtension on _LEVELS? {
  //String get name => this.toString().substring(this.toString().lastIndexOf('.'));

  int get intValue {
    switch(this) {
      case _LEVELS.FATAL: return 1;
      case _LEVELS.ERROR: return 2;
      case _LEVELS.WARN: return 3;
      case _LEVELS.INFO: return 4;
      case _LEVELS.DEBUG: return 5;
      default: return 6;
    }
  }

  bool operator <(_LEVELS other) => this.intValue < other.intValue;
  bool operator >(_LEVELS other) => this.intValue > other.intValue;
  bool operator <=(_LEVELS? other) => this.intValue <= other.intValue;
  bool operator >=(_LEVELS other) => this.intValue >= other.intValue;

  String ToString() {
    switch(this)  {
      case _LEVELS.FATAL: return 'FATAL';
      case _LEVELS.ERROR: return 'ERROR';
      case _LEVELS.WARN: return 'WARN';
      case _LEVELS.INFO: return 'INFO';
      case _LEVELS.DEBUG: return 'DEBUG';
      default: return 'TRACE';
    }
  }
}

class _LogHistory {
  static const MAX_MESSAGES = 5000;
  static const MAX_HISTORY_SIZE = 1024*1024*5

  List<LogTrace> messages = [];
  int messagesSize = 0;

  static _LogHistory? _instance;
  static _LogHistory get instance {
    if(_instance==null)
      _instance = _LogHistory();

    return _instance!;
  }

  void _addMessage(LogTrace message) {
    messages.add(message);
    messagesSize += message.size;
    clearOld(); 
  }
  
  void fatal(String m) {
    _addMessage(LogTrace(_LEVELS.FATAL,m));        
  }

  void error(String m) {
    _addMessage(LogTrace(_LEVELS.ERROR,m));        
  }

  void warn(String m) {
    _addMessage(LogTrace(_LEVELS.WARN,m));        
  }

  void info(String m) {
    _addMessage(LogTrace(_LEVELS.INFO,m));        
  }

  void debug(String m) {
    _addMessage(LogTrace(_LEVELS.DEBUG,m));        
  }

  void trace(String m) {
    _addMessage(LogTrace(_LEVELS.TRACE,m));        
  }

  void clearOld() {
    while(!messages.isEmpty && (messages.length > MAX_MESSAGES || messagesSize > MAX_HISTORY_SIZE)) {
	  messagesSize -= messages[0].size;
      messages.removeAt(0);
    }
  }
}

class LogTrace {
  final String message;
  final _LEVELS level;
  int get size => message.length + 2;

  LogTrace(this.level,this.message);
}